/*
styles.css
https://github.com/ddh0/easy-llama/
*/

html {
    background-color: #000000;
    color: #eeeeee;
    margin: 0px;
    padding: 0px;
    overflow: hidden;
}

body {
    background-color: #000000;
    color: #eeeeee;
    margin: 0px;
    padding: 0px;
    overflow: hidden;
}

#pageContainer {
    width: 100svw;
    height: 100svh;
    margin: 0px;
    border-color: transparent;
    border-width: 0px;
    padding: 0px;
    overflow: hidden;
    color: #eeeeee;
    font-family: Verdana, Geneva, Tahoma, sans-serif;
    font-size: 1rem;
    line-height: 1.6rem;
}

#conversationWindow {
    display: flex;
    flex-direction: column-reverse;
    background-color: #161616;
    width: 80svw;
    height: 85svh;
    float: left;
    margin: 0px;
    border-color: inherit;
    border-width: inherit;
    padding: inherit;
    overflow-y: auto;
    color: inherit;
    font-family: inherit;
    font-size: inherit;
    line-height: inherit;
}

#bottomPanelContainer {
    display: flex;
    flex-direction: row;
    width: 100svw;
    height: 15svh;
    background-color: #000000;
    border-color: #545454;
    border-top-style: solid;
    border-top-width: 5px;
    padding: 0px;
    margin: 0px;
    overflow: hidden;
}

#inputBox {
    border-right-width: 3px;
    border-top: none;
    border-bottom: none;
    border-left: none;
    border-color: #212121;
    background-color: #161616;
    width: calc(80svw - 1rem); /* fill width and account for padding */
    height: 15svh;
    margin: 0px;
    resize: none;
    padding: 0.5rem;
    overflow: auto;
    color: #eeeeee;
    font-family: Verdana, Geneva, Tahoma, sans-serif;
    font-size: 0.87rem;
    line-height: 1.6rem;
}

#submitArea {
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    background-color: #161616;
    height: 15svh;
    width: 20svw;
    padding: 0px;
    margin: 0px;
    overflow: hidden;
}

#panel {
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    background-color: inherit;
    float: right;
    width: 20svw;
    height: 85svh;
    margin: inherit;
    border-color: inherit;
    border-width: inherit;
    padding: inherit;
    overflow: inherit;
    color: inherit;
    font-family: inherit;
    font-size: inherit;
    line-height: inherit;
}

button {
    border-color: #555;
    border-style: solid;
    border-width: 1px;
    border-radius: 10px;
    margin-top: 1rem;
    margin-bottom: 1rem;
    padding: 0px;
    width: 15svw;
    height: 7svh;
    background-color: #4b724b;
    color: #eeeeee;
    font-family: Verdana, Geneva, Tahoma, sans-serif;
    font-size: 0.75rem;
    line-height: 1rem;
    overflow: clip;
}

button:hover {
    background-color: #355335;
    border-color: #838383;
    cursor: pointer;
}

button.copy-button {
    background-color: #000000bf;
    color: #bebebe;
    border-width: 0px;
    border-radius: 4px;
    margin-top: 0rem;
    margin-bottom: 0rem;
    width: 8rem;
    height: 1.5rem;
    font-family: Verdana, Geneva, Tahoma, sans-serif;
    font-size: 0.75rem;
    line-height: 1rem;
    overflow: clip;
}

button.copy-button:hover {
    background-color: #00000075;
    cursor: pointer;
}

h1 {
    line-height: 4rem;
}

h2 {
    line-height: 3rem;
}

pre {
    font-size: 85%; 
    line-height: 1.2rem;
    /* max-height: 30rem; */
    text-align: left;
    direction: ltr;
    border-radius: 1rem;
    padding: 0.65rem;
    white-space: pre;
    word-wrap: normal;
    tab-size: 4;
    background-color: #000000bf;
    color: #bebebe;
    overflow: auto;
    max-width: 100%;
    hyphens: none;
}

code:not(pre code) { /* inline code */
    background-color: #00000080;
    color: #bebebe;
    font-size: 85%;
    border-radius: 7px;
    padding: 3px;
    hyphens: none;
    /* white-space: pre; */
}

a {
    color: #a0c5ff;
}

a:visited {
    color: #a0c5ff;
}

.message {
    font-family: Verdana, Geneva, Tahoma, sans-serif;
    font-size: 0.87rem;
    /* overflow: auto; */
    width: calc(100% - 2rem);
    padding: 0.0rem 1rem;
    border-radius: 0px;
    color: #eeeeee;
}

.user-message {
    background-color: #214a2e;
    align-self: center;
}

.bot-message {
    background-color: #21434b;
    align-self: center;
}

/* these 4 rules do the same thing on different browsers */
textarea::placeholder {
    color: #404040;
}
textarea::-webkit-input-placeholder {
    color: #404040;
}
textarea::-moz-placeholder {
    color: #404040;
}
textarea:-ms-input-placeholder {
    color: #404040;
}

textarea:focus {
    outline: none;
}

.disabled-button {
    background-color: #404040 !important ;
    cursor: not-allowed !important;
}

.disabled-button:hover {
    background-color: #404040 !important ;
}

.red-button {
    background-color: #693333;
}

.red-button:hover {
    background-color: #4c2323;
}

#resetButton {
    background-color: #693333;
}

#resetButton:hover {
    background-color: #4c2323;
}

#removeButton {
    background-color: #696233;
}

#removeButton:hover {
    background-color: #4c4723;
}

#newBotMessageButton {
    background-color: #335e69;
}

#newBotMessageButton:hover {
    background-color: #23444c;
}

#swipeButton {
    background-color: #363369;
}

#swipeButton:hover {
    background-color: #25234c;
}

#fileUploadButton {
    background-color: #7b7f7b;
}

#fileUploadButton:hover {
    background-color: #5a5d5a;
}

#summarizeButton {
    background-color: #734c6f;
}

#summarizeButton:hover {
    background-color: #543651;
}

#submitButton {
    height: 10svh;
    width: 18svw;
    font-size: 0.9rem;
}

#pageContainer h1 {
    text-align: center;
    color: #eeeeee;
}

#samplerForm {
    display: flex;
    flex-direction: column;
    align-items: center;
    margin-top: 20px;
}

.form-row {
    display: flex;
    justify-content: space-between;
    width: 80%;
    margin-bottom: 20px;
}

.form-column {
    flex: 1;
    margin-right: 20px;
}

.form-column:last-child {
    margin-right: 0;
}

#samplerForm label {
    color: #eeeeee;
    margin-bottom: 5px;
}

#samplerForm input {
    width: 100%;
    margin-bottom: 10px;
    padding: 5px;
    border: 1px solid #555;
    border-radius: 4px;
    background-color: #161616;
    color: #eeeeee;
}

.button-container {
    display: flex;
    justify-content: space-between;
    width: 80%;
    margin-top: 20px;
}

.button-container button {
    margin-right: 10px;
    width: 20svw;
}

.button-container button:last-child {
    margin-right: 0;
}