# _version.py
# https://github.com/ddh0/easy-llama/

# define package version once here, and import it from here in each file
# for easy updates

from llama_cpp import __version__ as __llama_cpp_version__

__version__ = '0.1.116'
